import 'package:fluartx/fluartx.dart';
import 'package:fluartx/src/preconditions.dart';
import 'package:fluartx/src/transformers.dart';

typedef Predicate<T> = bool Function(T);

class Predicates {
  static Predicate<Object> alwaysTrue = (t) => true;
  static Predicate<Object> alwaysFalse = (t) => false;
  static Predicate<Object> notNull = (t) => t != null;
  static Predicate<Object> isNull = (t) => t == null;

  static Predicate<T> not<T>(@nonNull Predicate<T> predicate) {
    return (t) => !predicate(t);
  }

  static Predicate<T> and<T>(@nonNull Iterable<Predicate<T>> components) {
    var copied = _defensiveCopy(components);
    return (t) {
      for (var element in copied) {
        if (!element(t)) {
          return false;
        }
      }
      return true;
    };
  }

  static Predicate<T> or<T>(@nonNull Iterable<Predicate<T>> components) {
    var copied = _defensiveCopy(components);
    return (t) {
      for (var element in copied) {
        if (element(t)) {
          return true;
        }
      }
      return false;
    };
  }

  static Predicate<T> equalTo<T>(@nullable T target) {
    return (target == null) ? isNull : (t) => t == target;
  }

  static Predicate<Object> instanceOf<T>(@nullable Type type) {
    return (t) => t.runtimeType == type;
  }

  static Predicate<T> containsIn<T>(@nonNull Iterable<T> target) {
    return (t) => target.contains(t);
  }

  static Predicate<A> compose<A, B>(
      @nonNull Predicate<B> predicate, @nonNull Transformer<A, B> transformer) {
    return (t) => predicate(transformer(t));
  }

  static List<T> _defensiveCopy<T>(@nonNull Iterable<T> iterable) {
    List<T> list = <T>[];
    for (var element in iterable) {
      list.add(checkNotNull(element));
    }
    return list;
  }
}
