import 'package:fluartx/fluartx.dart';

class Result<T> {
  @nullable
  final Object value;

  bool get isSuccess => !(value is Failure);

  bool get isFailure => value is Failure;

  Result(this.value);

  @nullable
  T getOrNull() {
    if (isFailure) {
      return null;
    } else {
      return value as T;
    }
  }

  @nullable
  Exception exceptionOrNull() {
    if (value is Failure) {
      return (value as Failure).exception;
    } else {
      return null;
    }
  }

  @override
  String toString() {
    if (value is Failure) {
      return value.toString();
    } else {
      return "Success($value)";
    }
  }

  throwOnFailure() {
    if (value is Failure) {
      throw (value as Failure).exception;
    }
  }

  T getOrThrow() {
    throwOnFailure();
    return value as T;
  }

  T getOrElse(T onFailure(Exception exception)) {
    var exception = exceptionOrNull();
    if (exception == null) {
      return value as T;
    } else {
      return onFailure(exception);
    }
  }

  T getOrDefault(T defaultValue) {
    if (isFailure) return defaultValue;
    return value as T;
  }

  Result<R> map<R>(R transform(T value)) {
    if (isSuccess) {
      return Result.success(transform(value as T));
    } else {
      return Result(value);
    }
  }

  Result<R> mapCatching<R>(R transform(T value)) {
    if (isSuccess) {
      return runCatching(() => transform(value as T));
    } else {
      return Result(value);
    }
  }

  /// Returns the the result of [onSuccess] for encapsulated value if this instance represents [success][Result.isSuccess]
  /// or the result of [onFailure] function for encapsulated exception if it is [failure][Result.isFailure].
  /// Note, that an exception thrown by [onSuccess] or by [onFailure] function is rethrown by this function.
  R fold<R>(R onSuccess(T value), R onFailure(Exception exception)) {
    var exception = exceptionOrNull();
    if (exception == null) {
      return onSuccess(value as T);
    } else {
      return onFailure(exception);
    }
  }

  Result<T> onFailure(action(Exception exception)) {
    var exception = exceptionOrNull();
    if (exception != null) {
      action(exception);
    }
    return this;
  }

  Result<T> onSuccess(action(T value)) {
    if (isSuccess) {
      action(value as T);
    }
    return this;
  }

  static Result<T> success<T>(T value) => Result(value);

  static Result<T> failure<T>(Exception exception) =>
      Result(createFailure(exception));
}

class Failure {
  final Exception exception;

  Failure(this.exception);

  @override
  bool operator ==(other) {
    return other is Failure && exception == other.exception;
  }

  @override
  int get hashCode => exception.hashCode;

  @override
  String toString() {
    return "Failure($exception)";
  }
}

Object createFailure(Exception exception) => Failure(exception);

/// Calls the specified function [block] and returns its encapsulated result if invocation was successful,
/// catching and encapsulating any thrown exception as a failure.
Result<R> runCatching<R>(R block()) {
  try {
    return Result.success(block());
  } catch (e) {
    return Result.failure(e);
  }
}
